#include "../src/grind.h"
#include "gl2ps.h"
#include <pthread.h>

#include <stdio.h>
#include <string.h>


void writepsfile(int format, int sort, int options, int nbcol, char *filename){
  FILE *fp;
  int state = GL2PS_OVERFLOW, buffsize = 0;

  extern double colormap[10][3];

  fp = fopen(filename, "wb");
  if(!fp){
    gr_warning("Unable to open file %s for writing\n", filename);
    return;
  }

  printf("Saving image to file %s... ", filename);
  fflush(stdout);

  /*  change black and white color values */
  /*  postscript picture gives black lines on white background */
  colormap[0][0] = 1.;
  colormap[0][1] = 1.;
  colormap[0][2] = 1.;
  colormap[1][0] = 0.;
  colormap[1][1] = 0.;
  colormap[1][2] = 0.;

  /*  while buffer is too small to render image increase buff size` */
  while(state == GL2PS_OVERFLOW){
    buffsize += 1024*1024;
    if (GL2PS_ERROR == 
	gl2psBeginPage(filename, "grind", NULL, format, sort, options,
		       GL_RGBA, 0, NULL, nbcol, nbcol, nbcol, 
		       buffsize, fp, filename)) {
      printf("error opening postscript file, returning\n");
      return;
    }
    /*  draw the image in postscript */
    display();
    state = gl2psEndPage();
  }

  fclose(fp);

  /*  reset the black and white colors */
  colormap[0][0] = 0.;
  colormap[0][1] = 0.;
  colormap[0][2] = 0.;
  colormap[1][0] = 1.;
  colormap[1][1] = 1.;
  colormap[1][2] = 1.;

  if (state != GL2PS_SUCCESS)
    gr_warning("something went wrong writing postscript file?\n");
  else
    printf("Done!\n");
  fflush(stdout);
  return;
}


/*  all PSdump related issues */
pthread_mutex_t psdump_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t psdump_cond = PTHREAD_COND_INITIALIZER;
/*  flag to signal that the graphics stack should be dumped to a PS file */
static boolean psdump_flag = false;

/*  some psdump options */
static int psdump_format = 0;
static int psdump_opt = 0;
static char *psdump_filename = NULL;

int requestPSdump(int format, int opt, char *filename) {
  /*  check if graphics is up and running */
  if (!checkgraphics()) {
    return 0;
  }

  if (DEBUG) return 0;

  /* lock psdump mutex to set flag and options */
  pthread_mutex_lock(&psdump_lock);
  psdump_flag = true;
  /*  setting options here */
  psdump_format = format;
  psdump_opt = opt;
  psdump_filename = filename;
  while(psdump_flag)
    pthread_cond_wait(&psdump_cond, &psdump_lock);
  /*  unlock mutex */
  pthread_mutex_unlock(&psdump_lock);
  return 1;
}

int checkPSdump() {
  boolean worked = false;
  int nbcol = 0;


  if (DEBUG) return 0;


  /*  lock psdump mutex to check flag and (optionally) read options */
  pthread_mutex_lock(&psdump_lock);
  if (psdump_flag) {
    /*  write postscript file with given options */
    writepsfile(psdump_format, GL2PS_BSP_SORT, psdump_opt, nbcol, 
		psdump_filename);
    /*  signal waiting thread dump is done */
    pthread_cond_signal(&psdump_cond);
    /*  reset flag */
    psdump_flag = false;
    worked = true;
  }
  /*  unlock the stack */
  pthread_mutex_unlock(&psdump_lock);
  return worked;
}

int export( int argc, char *argv[], char **filename) {
  int i;
/*  all options for gl2ps: */
  static int format = GL2PS_EPS;
  /*  static int sort = GL2PS_BSP_SORT; */
  static int opt = GL2PS_OCCLUSION_CULL /*  remove hidden polygons from file */
    | GL2PS_BEST_ROOT /*  optimize BSP tree */
    /* | GL2PS_COMPRESS *//* generate compressed PDF and PS files (using zlib) */
    /* | GL2PS_NO_PS3_SHADING *//*  PS level3; 'shfill' operator not used */
    | GL2PS_USE_CURRENT_VIEWPORT /*  no more need to specify viewport */
    ;/* | GL2PS_SIMPLE_LINE_OFFSET; *//*  small offset in z-buffer; acts as anti-alias */
  char *t_ext;

  if (argc > 1) {
    if (*filename)
      free(*filename);
    *filename = ccalloc(strlen(argv[1])+5, "'filename' in Export"); /* .eps */
    strcpy(*filename, argv[1]);
    if ( (t_ext = strrchr(argv[1], '.')) ) {
      t_ext++;
      strcpy(*filename,argv[1]);
      if (0 == strcmp(t_ext, "pdf")) {
        format = GL2PS_PDF;
      }
      else if (0 == strcmp(t_ext, "ps")) {
        format = GL2PS_PS;
      }
      else if (0 == strcmp(t_ext, "eps")) {
        format = GL2PS_EPS;
      }
      else if (0 == strcmp(t_ext, "tex")) {
        format = GL2PS_TEX;
      }
    }
    else {
      format = GL2PS_EPS;
      strcat(*filename, ".eps");
    }
  }
  for (i=1; i < argc; i++) {
    if (eqf(argv[i], "notext")) {
      opt |= GL2PS_NO_TEXT;
    }
    else if (eqf(argv[i], "landscape")) {
      opt |= GL2PS_LANDSCAPE;
    }
  }
  if (GL2PS_PS == format)
    gps(*filename);
  else
    requestPSdump(format, opt, *filename);

  return 1;
}

