#include <stdio.h>
#include <math.h>

#include "grind.h"

int MAIN__(){return 1;}


extern boolean map_opt;

static double *wr = NULL, *wi = NULL;
double *eigvector = NULL;
int *eigsort = NULL, *eigstab = NULL, *eigcomp = NULL;

int eigen(double *pop, boolean matz) {
  static double *ddpop = NULL, *fv1 = NULL, tot;
  static int *iv1 = NULL;
  int ierr, i, j, n = nvar;


  if (!ddpop) ddpop = dcalloc(nvar*nvar, "'ddpop' in eigen()");
  if (!wr) wr = dcalloc(nvar, "'wr' in eigen()");
  if (!wi) wi = dcalloc(nvar, "'wi' in eigen()");
  if (!eigvector) eigvector = dcalloc(nvar*nvar, "'eigvector' in eigen()");
  if (!fv1) fv1 = dcalloc(nvar, "'fv1' in eigen()");

  if (!iv1) iv1 = icalloc(nvar, "'iv1' in eigen()");
  if (!eigsort) eigsort = icalloc(nvar, "'eigsort' in eigen()");
  if (!eigstab) eigstab = icalloc(nvar, "'eigstab'in eigen()");
  if (!eigcomp) eigcomp = icalloc(nvar, "'eigcomp'in eigen()");

  jac(pop, ddpop);
  if (map_opt)
    for (i=0; i<nvar; i++)
      ddpop[i*nvar + i] += 1.;
  /* rewriting 'ddpop' to fortran order; misusing 'eigvector' as temp-array  */
  for (i=0; i < nvar; i++)
    for (j=0; j < nvar; j++) 
      eigvector[j*nvar+i] = ddpop[i*nvar+j];
  for (i=0; i< nvar*nvar; i++) { 
    ddpop[i] = eigvector[i];
    eigvector[i] = 0.0;
  }
  rg_(&n, &n, ddpop, wr, wi, &matz, eigvector, iv1, fv1, &ierr);

  if (ierr) {
    printf("Error in computing eigenvalue: %d\n", ierr);
    return 0.0;
  }

  if (map_opt) {       /*  computing floquet multipliers */
    for (i=0; i<nvar; i++)
      wr[i] = sqrt((wr[i]*wr[i])+(wi[i]*wi[i]));
  }
  sort(nvar, wr, eigsort);  
  
  for (i=0; i<nvar; i++) {
    if (map_opt) {
      eigcomp[i] = 0;
      if (fabs(wr[i]-1.0) < TINY) eigstab[i] = NEUTRAL;
      else if (wr[i] < 1.0) eigstab[i] = STABLE;
      else eigstab[i] = UNSTABLE;
    }
    else {
      eigcomp[i] = (fabs(wi[i]) > TINY);  /*  complex eigenvalue */
      if (fabs(wr[i]) < TINY)
	eigstab[i] = NEUTRAL;    /*  bifurcation point */   
      else if (wr[i] < TINY)
	eigstab[i] = STABLE;    /*  real eigenvalue, stable point */
      else
	eigstab[i] = UNSTABLE;  /*  real eigenvalue, unstable point */
    }
  }
  if (matz) {
    for (i=0; i<nvar; i++) {
      tot = 0.;
      for (j=0; j<nvar; j++)
	tot += eigvector[i*nvar + j]*eigvector[i*nvar + j];
      tot = sqrt(tot);
      if (fabs(tot) > TINY)
	for (j=0; j<nvar; j++)
	  eigvector[i*nvar + j] /= tot;
    }
  }
  return eigstab[eigsort[0]];
}

int prieig(double pop[], boolean dojac) {
  int i, j;
  static double *ddpop = NULL;

  if (!ddpop) ddpop = dcalloc(nvar*nvar, "'ddpop' in prieig()");

  if (dojac) {
    printf("Jacobian matrix: \n");
    jac(pop,ddpop);
    if (map_opt)
      for (i=0; i<nvar; i++)
	ddpop[i*nvar + i] += 1.;
    for (i=0; i<nvar; i++) {
      for (j=0; j<nvar; j++) {
	printf(vformat, ddpop[i*nvar + j]);
	/* printf("\t"); */
      }
      printf("\n");
    }
    printf("\n");
  }
  if(map_opt)
    printf("floquet multipliers:\n");
  else {
    printf("    ");
    printf(sformat,"Lambda:");
    printf(sformat,"i/Vector:");
    printf("\n");
  }

  for (i=0; i<nvar; i++) {
    if (map_opt) {
      printf("%2d: ", i+1);
      printf(lformat, wr[eigsort[i]]);
      printf("\n");
    }
    else if (eigcomp[eigsort[i]]) {
      printf("%2d: ", i+1);
      printf(lformat, wr[eigsort[i]]);
      printf(" +i");
      printf(lformat, wi[eigsort[i]]);
      printf("\n");
    }
    else {
      printf("%2d: ", i+1);
      printf(lformat, wr[eigsort[i]]);
      printf(sformat, " Vector: [");
      for (j=0; j<nvar-1; j++) {
	printf(vformat, eigvector[eigsort[i]*nvar + j]);
	printf(", ");
      }
      printf(vformat, eigvector[eigsort[i]*nvar + nvar-1]);
      printf("]\n");
    }
  }
  return 1;
}
