#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "grind.h"

extern double finish;
extern int nintervals;
extern boolean adjust_opt;

static int poinc_dir = UP;
/* int poinc_points = 1; */
static double poinc_val = 0.5;
static double poinc_tol = 0.0001;
static int poinc_var = 0;

static double minval=0., maxval=0.;

int poincare(int points) {
  int ip, iter, max_iter = 20;
  double t,tend,dt,dx1,dx2,period,finint;
  static double *state = NULL;
  int nstep, istep;
  boolean orig_points_opt;

  if (Interrupt()) return 0;
  if (state == NULL) state = dcalloc(nvar, "'state' in poincare()");
  orig_points_opt = points_opt;
  points_opt = true;
  finint = finish/nintervals;
  period = 0.0;
  nstep = nintervals; 
  coparr(nvar, fstate, istate);
  for (ip = 0; ip < points; ip++) {
    t = 0;    
    dt = finint;
    /* iter = 0; */
    for (iter = istep = 0; istep < nstep  && !iter; istep++) {
      coparr(nvar, state, fstate);
      t = istep*finint;
      tend = (istep+1)*finint;
      onestep(t,tend,fstate);
      dx1 = poinc_val - state[poinc_var];
      dx2 = poinc_val - fstate[poinc_var];
      if (istep && dx1*dx2 < 0 && ((poinc_dir == BOTH) ||
	  (poinc_dir == DOWN && dx2 > 0) || (poinc_dir == UP && dx1 > 0))) {   
	/* t = tend - dt; */
	dt *= fabs(dx1)/fabs(fstate[poinc_var]-state[poinc_var]);
	coparr(nvar,fstate,state);
	for (iter = 1; iter <= max_iter && fabs(dx1) > poinc_tol; iter++) {
	  tend = t + dt;
	  onestep(t,tend,fstate);
	  dx1 = poinc_val - state[poinc_var];
	  dx2 = poinc_val - fstate[poinc_var];
	  if (dx1*dx2 > 0) {              /* Failed to cross poincare plane */
	    t = tend;
	    dt *= fabs(dx2)/fabs(fstate[poinc_var]-state[poinc_var]);
	    coparr(nvar, state, fstate);
	  } 
	  else {
	    if (fabs(dx2) < poinc_tol) {
	      /* coparr(nvar, state, fstate); */
	      break;
	    }
	    /* t = tend - dt; */
	    dt *= .9;
	    coparr(nvar, fstate, state);
	  }
	  /* t = tend; */
	}
	if (iter == max_iter) {
	  gr_warning("Poincare failed to converge after %d iterations\n",iter);
	  points_opt = orig_points_opt;
	  return 0;
	}
	start_trajectory(WHITE);
	if (ip) period += tend;
      }
    }
    if (istep >= nstep) {
      gr_warning("Poincare never crossed the plane at %s=%f\n",all_names[poinc_var],poinc_val);
      points_opt = orig_points_opt;
      return 0;
    }
  }
  if (points-1 && poinc_dir != BOTH)
    printf("Average period: %f\n",period/(points-1));
  points_opt = orig_points_opt;

  return 1;
}

int Poincare(int argc, char *argv[]) {
  static int points = 1;  /*  no. of times poincare crosses are plotted */
  int t;
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp
      ("poincare var real [up/down/both]  Or poincare npoints tolerance",
       " ","Set up a poincare plane  Or  Plot poincare crossings");
  if (eqArg(1, "status", argc, argv))
    return printf("Status poincare: var=%s real=%.3g direction=%s  Or  npoints=%d tolerance=%.3g\n",
		  all_names[poinc_var], poinc_val, 
		  (UP==poinc_dir ? "up" : (DOWN==poinc_dir ? "down": "both")),
		  points, poinc_tol);
  
  if (1 == argc)
    return 0;
  t = nameIndex(argv[1]);
  if (t >= 0 && t< nvar) {
    poinc_var = t;
    if (isnumeric(2, argc, argv))
      poinc_val = atof(argv[2]);
    else return gr_warning("read error\n");
    if (argc>3) {
      if (eqArg(3, "up", argc, argv))
	poinc_dir = UP;
      else if (eqArg(3, "down", argc, argv))
	poinc_dir = DOWN;
      else if (eqArg(3, "both", argc, argv))
	poinc_dir = BOTH;
      else
	gr_warning("read error\n");
    }
  }
  else {
    if (argc>1) {
      if (isnumeric(1, argc, argv))
	points = atoi(argv[1]);
      else return gr_warning("read error\n");
    }
    if (argc>2) {
      if (isnumeric(2, argc, argv))
	poinc_tol = atof(argv[2]);
      else return gr_warning("read error\n");
    }
    return poincare(points);
  }
  return 0;
}

int bifurcate(int ax, int steps, double stepsize, 
	      double transient, int points) {
  int i, j;
  boolean okay = true;
  double finint, t;
  static double *pop = NULL;
  boolean orig_points_opt;
  double orig_poinc_val;
  double startval;   /* startval is value from which we start bifurcation */
  double parval;

  if (Interrupt()) return 0;

  parval = all_values[axis[ax].var];
  if (parval < axis[ax].min || parval > axis[ax].max) 
    gr_warning("Initial parameter value not in diagram: %s=%f\n",
	       all_names[axis[ax].var],parval); 


  if (NULL == pop) pop = dcalloc(nvar, "'pop' in bifurcate()");
  finint = finish/nintervals;
  /* store original values */
  storeaxis(true);
  orig_points_opt = points_opt;
  orig_poinc_val = poinc_val;
  /* also store istate which we're changing below 
     in order for poincare() to work properly */
  coparr(nvar, pop, istate);
  /* copy istate to the working vector fstate */
  coparr(nvar, fstate, istate);

  if (adjust_opt) /*  initial transient to get into the periodic regime 
		      plus initialization of min- and max-vals*/
    okay = onestep(0.0,transient,fstate);

  startval = val2ax(all_values[axis[ax].var], ax);
  for (i=0; i<=steps && okay; i++) {
    all_values[axis[ax].var] = ax2val(startval + i*stepsize, ax);
    if (false == keepvar_opt) /* copy original istate to fstate */
      coparr(nvar, fstate, pop);
    if (adjust_opt) {
      minval = maxval = poinc_val;
      t = 0.;
      while (t<transient+TINY && (okay = onestep(t, t+finint, fstate))) {
	if (fstate[poinc_var] < minval)
	  minval = fstate[poinc_var];
	if (fstate[poinc_var] > maxval)
	  maxval = fstate[poinc_var];
	t+=finint;
      }
      poinc_val = (maxval-minval)/2.0 + minval;
    }
    else
      okay = onestep(0.0, transient, fstate);
    if (!okay) break;
    if (map_opt) {
      points_opt = true;
      start_trajectory(WHITE);
      for (j=0; j<points; j++) {
	onestep(0.0, 1.0, fstate);
	add2trajectory();
      }      
    }

    else {
      coparr(nvar, istate, fstate);
      if (!poincare(points))
	break;
    }
  }
  points_opt = orig_points_opt;
  poinc_val = orig_poinc_val;
  storeaxis(false);
  coparr(nvar, istate, pop);
  return okay;
}

int Bifurcate(int argc, char *argv[]) {
  static int ax = X_AXIS;
  static double transient = 100.;
  static int steps = 50;
  static int points = 1;
  int iax, upsteps, downsteps;
  double stepsize;

  /*  check first whether help or status is asked for */
  if ( eqArg(1, "help", argc, argv))
    return gHelp("bifurcate x npoints nsteps transient",
		 "bifurcate x 1 50 100.0",
		 "Bifurcate continues a parameter plotting poincare sections");

  if (eqArg(1, "status", argc, argv))
    return printf("Status bifurcate: axis=%s npoint=%d nsteps=%d transient=%.3g\n",
		  axis[ax].label, points, steps, transient);

  if (argc>1) {
    iax = Index("xyz", argv[1]);
    if (iax < X_AXIS || iax > Z_AXIS || axis[iax].var < nvar+nexpr) {
      gr_warning("not an axis: %s, or not an parameter: %s\n", 
		 argv[1], axis[iax].label);
      return 0;
    }
    else
      ax = iax;
  }
  if (argc>2) {
    if (isnumeric(2, argc, argv))
      points = atoi(argv[2]);
    else
      return gr_warning("read error\n");
  }
  if (argc>3) {
    if (isnumeric(3, argc, argv))
      steps = atoi(argv[3]);
    else return gr_warning("read error\n");
  }
  if (argc>4) {
    if (isnumeric(4, argc, argv))
      transient = atof(argv[4]);
    else return gr_warning("read error\n");
  }
  else
    transient = finish;

  downsteps = val2ax(all_values[axis[ax].var], ax)*steps + .5;
  upsteps = steps - downsteps;
  stepsize = 1.0/steps;
  return bifurcate(ax, upsteps, stepsize, transient, points) &&
    bifurcate(ax, downsteps, -stepsize, transient, points);
}
